import logging

from mesa import Agent

import elecsim.scenario.scenario_data
logger = logging.getLogger(__name__)



"""demand.py: Agent which simulates the demand of the UK"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "Alexander@Kell.es"


class Demand(Agent):

    def __init__(self, model, unique_id, segment_hours=None, segment_consumption=None):
        """ An agent representing UK electricity demand

        :param segment_hours: A series representing the load duration curve
        """
        self.unique_id = unique_id
        self.model = model
        self.segment_hours = segment_hours
        self.segment_consumption = segment_consumption

        self.year_segment_hours = segment_hours
        self.year_segment_consumption = segment_consumption

        self.yearly_demand_change = elecsim.scenario.scenario_data.yearly_demand_change
        self.years_from_start = 0

    def step(self):
        logger.debug("Stepping demand")
        self.segment_consumption = [i * self.yearly_demand_change[self.years_from_start] for i in self.segment_consumption]
        self.years_from_start += 1

        # load to change each year due to certain scenario




import os
from functools import reduce
from itertools import count, islice, cycle

import numpy as np
import pandas as pd
from pandas.api.types import is_string_dtype

pd.set_option('display.max_columns', 70)
pd.set_option('display.max_rows', 200)


def plant_data(dat):
    """
    Takes the first part of the dataframe from the CSV file from the tables included in BEIS_Electricity_Generation_Cost_Report, and returns a dataframe with powerplants as columns and each data as columns
    :param dat: Input dataframe from BEIS_Electricity_Generation_Cost_Report
    :return: Returns first part of table, which includes plant size, average load factor, efficiency and operating period with power plants as rows and columns as data about said power plants
    """

    characteristics = dat.loc[0:2].dropna(axis='columns').drop(['Unnamed: 1'], axis=1).transpose()  # Fills empty column names, deletes units, and transposes dataframe so that plants are as rows
    operating_period = dat.loc[9].dropna().drop(['Unnamed: 1']).transpose().to_frame()  # Transposes operating period and drops units

    dat = characteristics.merge(operating_period, left_index=True, right_index=True)  # Merges the two series together to form a single dataframe

    dat = dat.rename(index=str, columns = {0:"Plant_Size",1:"Average_Load_Factor",2:"Efficiency",9:"Operating_Period"})  # Rename columns
    return dat


def development_period(dat):
    """
    Takes the entire dataframe from a table included in BEIS_Electricity_Generation_Cost_Report and returns data about the pre-development and construction cost expenditure information per year
    :param dat: Dataframe of the table included in BEIS_Electricity_Generation_Cost_Report
    :return: Returns a dataframe of power plant data which contains pre-development and construction expenditure information per year
    """
    dat = dat.iloc[3:9].copy()
    dat.iloc[:, 0] = dat.iloc[:, 0].fillna(method='ffill')

    dat.iloc[:, 0] = dat.iloc[:, 0].replace({  # Shorten column names
                                                    "Pre-development period ":"Pre", "Construction period ":"Constr"})

    dat.iloc[:, 1] = dat.iloc[:, 1].replace({  # Shorten column names
                                                    "Duration and % spend per years 1 & 2 ": "Dur,1,2",
                                                    "% spend per years 3, 4, & 5 ": "3,4,5",
                                                    "% spend per years 6, 7, & 8 ": "6,7,8"})

    dat['A'], dat['B'], dat['C'] = dat.iloc[:, 1].str.split(',', 2).str  # Split duration and spend per year to respective column
    dat = dat.drop(['Unnamed: 1'], axis=1)
    dat['A'], dat['B'], dat['C'] = dat['Unnamed: 0'] + "_" + dat['A'], dat['Unnamed: 0'] + "_" + dat['B'], dat['Unnamed: 0'] + "_" + dat['C']  # Merge column name with year information

    tot_cols = []
    for it in islice(count(), 1, len(dat.columns), 3):  # Skip through to columns containing power plant information
        cycle_period = cycle(['A', 'B', 'C'])  # Cycle through rows to create column names
        for val in range(0, 3):  # Three values per power plant
            period = next(cycle_period)
            cols = pd.DataFrame({'plant': dat.columns.values[it],'period': dat[period],'value': dat.iloc[:, it+val]})  # Create dataframe in wide format
            tot_cols.append(cols)
    db = pd.concat(tot_cols).reset_index().drop(['index'],axis=1)  # Concatenate all dataframes
    db = db.pivot(index='plant', columns='period', values='value')  # Transform dataframe to wide format

    return db


def costs(dat):
    """
    Takes dataframe from CSV file of BEIS_Electricity_Generation_Cost_Report and returns cost section of data into a wide format with power plant data per row
    :param dat: Dataframe starting from Table 19 found in BEIS_Electricity_Generation_Cost_Report
    :return: Returns dataframe of cost information in wide format
    """
    dat=dat[10:].copy()
    dat.iloc[:, 0] = dat.iloc[:, 0].fillna(method='ffill')  # Fill in blanks of power plant information

    dat.iloc[:, 0] = dat.iloc[:, 0].replace({  # Shorten column names
                                                    "Pre-development £/kW ":"Pre_dev_cost", "Construction £/kW ":"Constr_cost",
                                                    "Infrastructure £'000s ":"Infra_cost", "Fixed O&M £/MW/year ":"Fixed_cost",
                                                    "Variable O&M £/MWh ":"Var_cost","Insurance £/MWh/year ":"Insurance_cost",
                                                    "Connection and Use of System charges £/MW/year ":"Connect_system_cost"
                                                    })
    dat.iloc[:, 0] = dat.iloc[:, 0] + "-" + dat.iloc[:, 1]  # Merge cost information with projection (high, medium or low)
    dat = dat.drop(['Unnamed: 1'], axis=1)

    col_names = []
    col_values = []
    plant_names = [x for z in range(13) for x in dat.columns if not "Unnamed:" in x for y in range(3)]  # Create list containing plant names in correct order

    for x in dat[2:].itertuples(index=False):  # Loop through each row of dataframe
        for j in range(1,len(x)):
            col_names.append(x[0]+"_"+dat.iloc[1, j])  # Create column names by merging year of cost with cost type and projection
            col_values.append(x[j])  # Take column value

    cols_df = pd.DataFrame({'columns':col_names,'plant':plant_names,'col_values':col_values})  # Create dataframe
    cols_df = cols_df.pivot(index="plant",values='col_values',columns='columns')  # Make into wide format
    return cols_df


def merge_plant_cost_tables(df):
    """
    Function to bring together all sections of dataframe
    :param df: Dataframe starting from Table 19 found in BEIS_Electricity_Generation_Cost_Report
    :return: Returns final dataframe found in BEIS_Electricity_Generation_Cost_Report Table 19 in wide format
    """

    plant = plant_data(df)  # Call functions
    dev = development_period(df)
    cost = costs(df)

    dfs = [plant, dev, cost]  # Create a list of dataframes

    df_final = reduce(lambda left, right: pd.merge(left,right, left_index=True, right_index=True), dfs)  # Merge each dataframe together
    return df_final


def beis_human_to_machine_format(dir):

    final_df = []
    for file in os.listdir(dir):
        file_p = "/Users/b1017579/Documents/PhD/Projects/10. ELECSIM/10. ELECSIM/data/Power_Plants/Power_Plant_costs/plant_costs_files/"+file
        df = pd.read_csv(file_p).replace("- ",np.nan)
        final_df.append(merge_plant_cost_tables(df))
    final_df = pd.concat(final_df)

    # Remove comma seperator for increments of 1000 and unit of year
    final_df = final_df.apply(lambda x: x.str.replace(',', ''), axis=1).apply(lambda x: x.str.replace(' years',''), axis=1)
    # Remove the % sign and divide by 100
    final_df = final_df.apply(lambda x: x.str.replace("%", "").astype('float')/100 if is_string_dtype(x) and x.str.contains('%').all() else x)
    print(final_df)
    return final_df


p_dat = beis_human_to_machine_format("/Users/b1017579/Documents/PhD/Projects/10. ELECSIM/10. ELECSIM/data/Power_Plants/Power_Plant_costs/plant_costs_files")


p_dat.to_csv("/Users/b1017579/Documents/PhD/Projects/10. ELECSIM/10. ELECSIM/data/Power_Plants/Power_Plant_costs/plant_cost_data_nan.csv")




import numpy as np
"""
File name: world_environment
Date created: 19/02/2019
Feature: # Creation of Environment for Reinforcement Learning
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"



from functools import lru_cache

from elecsim.plants.plant_type.power_plant import PowerPlant
from elecsim.role.plants.costs.non_fuel_cost_calculations import NonFuelCostCalculation

""" no_fuel_plant.py: Child class of power plant which contains functions for a power plant that does not require fuel"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "Alexander@Kell.es"


class NonFuelPlant(PowerPlant):

    def __init__(self, name, plant_type, construction_year, capacity_mw, average_load_factor, pre_dev_period, construction_period, operating_period, pre_dev_spend_years, construction_spend_years, pre_dev_cost_per_mw, construction_cost_per_mw, infrastructure, fixed_o_and_m_per_mw, variable_o_and_m_per_mwh, insurance_cost_per_mw, connection_cost_per_mw, efficiency):
        """
        Power plant of type that does not use plant_type.
        """
        super().__init__(name, plant_type, capacity_mw, construction_year, average_load_factor, pre_dev_period, construction_period, operating_period, pre_dev_spend_years, construction_spend_years, pre_dev_cost_per_mw, construction_cost_per_mw, infrastructure, fixed_o_and_m_per_mw, variable_o_and_m_per_mwh, insurance_cost_per_mw, connection_cost_per_mw)

        self.efficiency = efficiency
        self.min_running = 0

    def calculate_lcoe(self, discount_rate):
        lcoe_object = NonFuelCostCalculation(capacity_mw=self.capacity_mw, construction_year=self.construction_year, average_load_factor=self.average_load_factor, efficiency=self.efficiency, pre_dev_period=self.pre_dev_period, construction_period=self.construction_period, operating_period=self.operating_period, pre_dev_spend_years=self.pre_dev_spend_years, construction_spend_years=self.construction_spend_years, pre_dev_cost_per_mw=self.pre_dev_cost_per_mw, construction_cost_per_mw=self.construction_cost_per_mw, infrastructure=self.infrastructure, fixed_o_and_m_per_mw=self.fixed_o_and_m_per_mw, variable_o_and_m_per_mwh=self.variable_o_and_m_per_mwh, insurance_cost_per_mw=self.insurance_cost_per_mw, connection_cost_per_mw=self.connection_cost_per_mw)
        lcoe = lcoe_object.calculate_lcoe(discount_rate)
        return lcoe

    @lru_cache(maxsize=1024)
    def short_run_marginal_cost(self, model, genco, fuel_price=None, co2_price=None):
        plant_cost_calculations = NonFuelCostCalculation(capacity_mw=self.capacity_mw, construction_year=self.construction_year, average_load_factor=self.average_load_factor, efficiency=self.efficiency, pre_dev_period=self.pre_dev_period, construction_period=self.construction_period, operating_period=self.operating_period, pre_dev_spend_years=self.pre_dev_spend_years, construction_spend_years=self.construction_spend_years, pre_dev_cost_per_mw=self.pre_dev_cost_per_mw, construction_cost_per_mw=self.construction_cost_per_mw, infrastructure=self.infrastructure, fixed_o_and_m_per_mw=self.fixed_o_and_m_per_mw, variable_o_and_m_per_mwh=self.variable_o_and_m_per_mwh, insurance_cost_per_mw=self.insurance_cost_per_mw, connection_cost_per_mw=self.connection_cost_per_mw)
        marginal_cost = plant_cost_calculations.calculate_short_run_marginal_cost(model)
        return marginal_cost



import logging

from scipy.optimize import minimize

import elecsim.scenario.scenario_data
from elecsim.plants.fuel.fuel_registry.fuel_registry import plant_type_to_fuel
from elecsim.plants.plant_costs.estimate_costs.estimate_old_plant_cost_params.old_plant_param_calc import OldPlantCosts
from elecsim.plants.plant_type.fuel_plant import FuelPlant

logger = logging.getLogger(__name__)



class FuelOldPlantCosts(OldPlantCosts):

    historic_fuel_price = elecsim.scenario.scenario_data.historical_fuel_prices_long

    def __init__(self, year, plant_type, capacity):
        super().__init__(year=year, plant_type=plant_type, capacity=capacity)
        self.fuel = plant_type_to_fuel(self.plant_type, self.year)

    def estimate_cost_parameters(self):
        """
        Function which estimates the parameters of the non-modern fuel power plant using the LCOE value for the relevant
        year.
        :return: Returns dictionary of the updated parameters for the power plant.
        """

        params_for_scaling = self.estimate_modern_parameters()

        parameter_values = list(params_for_scaling.values())

        linear_optimisation_results = self._linear_optimisation(parameter_values, self.estimated_historical_lcoe)
        linear_optimisation_parameters = linear_optimisation_results['x'].tolist()

        scaled_parameters = {key: params for key, params in
                             zip(self.estimated_modern_plant_parameters, linear_optimisation_parameters)
                             if key not in self.params_to_ignore}

        # logger.debug("scaled_params: {}".format(scaled_parameters))
        # scaled_parameters['variable_o_and_m_per_mwh'] += uniform(-20,20)
        # scaled_parameters['variable_o_and_m_per_mwh'] *= uniform(0.3,2)
        scaled_parameters.update(self.dict_to_ignore)
        return scaled_parameters

    def estimate_modern_parameters(self):
        # self.params_to_ignore = ['pre_dev_period', 'operating_period', 'construction_period', 'efficiency',
        #                          'average_load_factor', 'construction_spend_years', 'pre_dev_spend_years']
        # self.dict_to_ignore = {key: self.estimated_modern_plant_parameters[key] for key in
        #                        self.estimated_modern_plant_parameters
        #                        if key in self.params_to_ignore}

        self.get_params_to_scale()

        params_for_scaling = {key: self.estimated_modern_plant_parameters[key] for key in
                              self.estimated_modern_plant_parameters
                              if key not in self.dict_to_ignore}
        return params_for_scaling

    def _linear_optimisation(self, x, lcoe_required):

        # x = [0.00001 if param == 0 else param for param in x]
        connection_cost_per_mw = x[0]
        construction_cost_per_mw = x[1]
        fixed_o_and_m_per_mw = x[2]
        infrastructure = x[3]
        insurance_cost_per_mw = x[4]
        pre_dev_cost_per_mw = x[5]
        variable_o_and_m_per_mwh = x[6]
        cons = [{'type': 'eq', 'fun': lambda x: x[0] / x[1] - connection_cost_per_mw / construction_cost_per_mw},
                {'type': 'eq', 'fun': lambda x: x[1] / x[2] - construction_cost_per_mw / fixed_o_and_m_per_mw},
                {'type': 'eq', 'fun': lambda x: x[2] / x[3] - fixed_o_and_m_per_mw / infrastructure},
                {'type': 'eq', 'fun': lambda x: x[3] / x[4] - infrastructure / insurance_cost_per_mw},
                {'type': 'eq', 'fun': lambda x: x[4] / x[5] - insurance_cost_per_mw / pre_dev_cost_per_mw},
                {'type': 'eq', 'fun': lambda x: x[5] / x[6] - pre_dev_cost_per_mw / variable_o_and_m_per_mwh},
                {'type': 'ineq', 'fun': lambda x: x[0]},
                {'type': 'ineq', 'fun': lambda x: x[1]},
                {'type': 'ineq', 'fun': lambda x: x[2]},
                {'type': 'ineq', 'fun': lambda x: x[3]},
                {'type': 'ineq', 'fun': lambda x: x[4]},
                {'type': 'ineq', 'fun': lambda x: x[5]},
                {'type': 'eq', 'fun': lambda x: self._calculate_lcoe_wrapper(x)-lcoe_required}]
        result = minimize(self._calculate_lcoe_wrapper, x0=x, constraints=cons)
        return result

    def _calculate_lcoe_wrapper(self, x0):
        """
        Calculates the LCOE class through the use of the FuelPlant class.
        :param params: Dict which contains parameters for the FuelPlant class
        :return: LCOE
        """
        test_params = {key:value for key, value in zip(self.estimated_modern_plant_parameters, x0)}

        holder_plant = FuelPlant(**test_params, name="LinOptim", plant_type=self.plant_type, capacity_mw=self.capacity,
                                 construction_year=self.plant.construction_year,
                                 average_load_factor=self.plant.average_load_factor,
                                 efficiency=self.plant.efficiency, pre_dev_period=self.plant.pre_dev_period,
                                 construction_period=self.plant.construction_period,
                                 operating_period=self.plant.operating_period,
                                 pre_dev_spend_years=self.plant.pre_dev_spend_years,
                                 construction_spend_years=self.plant.construction_spend_years,
                                 )
        lcoe = holder_plant.calculate_lcoe(self.discount_rate)

        return lcoe







#
# class FuelOldPlantCosts(OldPlantCosts):
#
#     historic_fuel_price = scenario.historical_fuel_prices_long
#
#     def __init__(self, year, plant_type, capacity):
#         super().__init__(year=year, plant_type=plant_type, capacity=capacity)
#         self.fuel = plant_type_to_fuel(self.plant_type, self.year)
#
#     def estimate_cost_parameters(self):
#         """
#         Function which estimates the parameters of the non-modern fuel power plant using the LCOE value for the relevant
#         year.
#         :return: Returns dictionary of the updated parameters for the power plant.
#         """
#
#         params_for_scaling = self.estimate_modern_parameters()
#
#         parameter_values = list(params_for_scaling.values())
#
#         linear_optimisation_results = self._linear_optimisation(parameter_values, self.estimated_historical_lcoe)
#         linear_optimisation_parameters = linear_optimisation_results['x'].tolist()
#
#         scaled_parameters = {key: params for key, params in
#                              zip(self.estimated_modern_plant_parameters, linear_optimisation_parameters)
#                              if key not in self.params_to_ignore}
#
#         # logger.debug("scaled_params: {}".format(scaled_parameters))
#         # scaled_parameters['variable_o_and_m_per_mwh'] += uniform(-20,20)
#         # scaled_parameters['variable_o_and_m_per_mwh'] *= uniform(0.3,2)
#         scaled_parameters.update(self.dict_to_ignore)
#         return scaled_parameters
#
#     def estimate_modern_parameters(self):
#         # self.params_to_ignore = ['pre_dev_period', 'operating_period', 'construction_period', 'efficiency',
#         #                          'average_load_factor', 'construction_spend_years', 'pre_dev_spend_years']
#         # self.dict_to_ignore = {key: self.estimated_modern_plant_parameters[key] for key in
#         #                        self.estimated_modern_plant_parameters
#         #                        if key in self.params_to_ignore}
#
#         self.get_params_to_scale()
#
#         params_for_scaling = {key: self.estimated_modern_plant_parameters[key] for key in
#                               self.estimated_modern_plant_parameters
#                               if key not in self.dict_to_ignore}
#         return params_for_scaling
#
#     def _linear_optimisation(self, x, lcoe_required):
#         connection_cost_per_mw = x[0]
#         construction_cost_per_mw = x[1]
#         fixed_o_and_m_per_mw = x[2]
#         infrastructure = x[3]
#         insurance_cost_per_mw = x[4]
#         pre_dev_cost_per_mw = x[5]
#         variable_o_and_m_per_mwh = x[6]
#         cons = [{'type': 'eq', 'fun': lambda x: x[0] / x[1] - connection_cost_per_mw / construction_cost_per_mw},
#                 {'type': 'eq', 'fun': lambda x: x[1] / x[2] - construction_cost_per_mw / fixed_o_and_m_per_mw},
#                 {'type': 'eq', 'fun': lambda x: x[2] / x[3] - fixed_o_and_m_per_mw / infrastructure},
#                 {'type': 'eq', 'fun': lambda x: x[3] / x[4] - infrastructure / insurance_cost_per_mw},
#                 {'type': 'eq', 'fun': lambda x: x[4] / x[5] - insurance_cost_per_mw / pre_dev_cost_per_mw},
#                 {'type': 'eq', 'fun': lambda x: x[5] / x[6] - pre_dev_cost_per_mw / variable_o_and_m_per_mwh},
#                 {'type': 'ineq', 'fun': lambda x: x[0]},
#                 {'type': 'ineq', 'fun': lambda x: x[1]},
#                 {'type': 'ineq', 'fun': lambda x: x[2]},
#                 {'type': 'ineq', 'fun': lambda x: x[3]},
#                 {'type': 'ineq', 'fun': lambda x: x[4]},
#                 {'type': 'ineq', 'fun': lambda x: x[5]},
#                 {'type': 'eq', 'fun': lambda x: self._calculate_lcoe_wrapper(x)-lcoe_required}]
#         result = minimize(self._calculate_lcoe_wrapper, x0=x, constraints=cons)
#         return result
#
#     def _calculate_lcoe_wrapper(self, x0):
#         """
#         Calculates the LCOE class through the use of the FuelPlant class.
#         :param params: Dict which contains parameters for the FuelPlant class
#         :return: LCOE
#         """
#         test_params = {key:value for key, value in zip(self.estimated_modern_plant_parameters, x0)}
#
#         holder_plant = FuelPlant(**test_params, name="LinOptim", plant_type=self.plant_type, capacity_mw=self.capacity,
#                                  construction_year=self.plant.construction_year,
#                                  average_load_factor=self.plant.average_load_factor,
#                                  efficiency=self.plant.efficiency, pre_dev_period=self.plant.pre_dev_period,
#                                  construction_period=self.plant.construction_period,
#                                  operating_period=self.plant.operating_period,
#                                  pre_dev_spend_years=self.plant.pre_dev_spend_years,
#                                  construction_spend_years=self.plant.construction_spend_years,
#                                  )
#         lcoe = holder_plant.calculate_lcoe(self.discount_rate)
#
#         return lcoe
#



from elecsim.plants.plant_type.fuel_plant import FuelPlant
from elecsim.plants.plant_type.non_fuel_plant import NonFuelPlant


class PlantRegistry:

    def __init__(self, plant_type):
        self.plant_type = plant_type

    def plant_type_to_plant_object(self):
        requires_fuel = self.check_if_fuel_required()
        return self._plant_object_registry(requires_fuel)

    def _plant_object_registry(self, requires_fuel):
        """
        Power Plant Registry which takes a boolean variable on whether a plant requires plant_type and returns object of plant_type
        either FuelPlant or RenewablePlant

        :param requires_fuel: Boolean on whether the power plant requires plant_type or not
        :return: Object of type of power plant
        """

        # requires_fuel = self.check_if_fuel_required()

        if not isinstance(requires_fuel, bool):
            # print("is not bool")
            raise ValueError("Must enter a boolean for plant_registry function.")

        if not requires_fuel:
            return NonFuelPlant
        elif requires_fuel:
            return FuelPlant


    def check_if_fuel_required(self):
        """
        Takes a plant_type type and returns a boolean specifying whether the power plant uses fuel or not.
        :param plant_type: Type of plant
        :return: Boolean specifying whether plant requires plant_type or not
        """
        plant_type = self.plant_type.lower()
        if plant_type == "gas":
            return True
        elif plant_type == "ccgt":
            return True
        elif plant_type == "ccgt":
            return True
        elif plant_type == "coal":
            return True
        elif plant_type == "pv":
            return False
        elif plant_type == "ad":
            return True
        elif plant_type == "act":
            return True
        elif plant_type == "offshore":
            return False
        elif plant_type == "biomass_wood":
            return True
        elif plant_type == "biomass_poultry_litter":
            return True
        elif plant_type == "biomass_straw":
            return True
        elif plant_type == "biomass_meat":
            return True
        elif plant_type == "onshore":
            return False
        elif plant_type == "landfill":
            return True
        elif plant_type == "sewage":
            return True
        elif plant_type == "geothermal":
            return False
        elif plant_type == "hydro_store":
            return False
        elif plant_type == "hydro":
            return False
        elif plant_type == "wave":
            return False
        elif plant_type == "tidal":
            return False
        elif plant_type == "ocgt":
            return True
        elif plant_type == "nuclear":
            return True
        elif plant_type == "recip_gas":
            return True
        elif plant_type == "pumped_storage":
            return False
        elif plant_type == "recip_diesel":
            return True
        elif plant_type == "efw":
            return True
        else:
            raise ValueError("Plant Type not Found: "+plant_type)

